{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "# https://github.com/stevenhalim/cpbook-code/blob/master/ch8/maxflow.cpp\n",
    "# https://github.com/stevenhalim/cpbook-code/blob/master/ch9/mcmf.cpp\n",
    "import sys \n",
    "INF = 1e18"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 40,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "125\n",
      "60\n"
     ]
    }
   ],
   "source": [
    "class graph:\n",
    "    class edge:\n",
    "        def __init__(self, node, cap, flow):\n",
    "            self.node, self.cap, self.flow = node, cap, flow\n",
    "        def __repr__(self):\n",
    "            return f'{self.node}:{self.cap}@{self.flow}'\n",
    "    def __init__(self, size):\n",
    "        self.nodes = [[] for _ in range(size)]\n",
    "        self.edges = []\n",
    "    def add(self, u, v, weight, directed = True):\n",
    "        if u == v: return\n",
    "        self.edges.append(graph.edge(v, weight, 0))\n",
    "        self.nodes[u].append(len(self.edges) - 1)\n",
    "        self.edges.append(graph.edge(u, weight if not directed else 0, 0))\n",
    "        self.nodes[v].append(len(self.edges) - 1)\n",
    "\n",
    "    # Find augmenting paths\n",
    "    def bfs(self, source, sink):\n",
    "        self.dist = [-1] * len(self.nodes)\n",
    "        self.prev = [(-1, -1)] * len(self.nodes)\n",
    "        queue = [source]\n",
    "        self.dist[source] = 0\n",
    "        while queue:\n",
    "            node = queue.pop(0)\n",
    "            if node == sink: break\n",
    "            for ei in self.nodes[node]:\n",
    "                e = self.edges[ei]\n",
    "                if e.cap - e.flow > 0 and self.dist[e.node] == -1:\n",
    "                    self.dist[e.node] = self.dist[node] + 1\n",
    "                    queue.append(e.node)\n",
    "                    self.prev[e.node] = (node, ei)\n",
    "        return self.dist[sink] != -1\n",
    "    \n",
    "    # Edmonds-Karp algorithm\n",
    "    def send_flow(self, source, sink, f = INF):\n",
    "        if source == sink: return f\n",
    "        node, ei = self.prev[sink]\n",
    "        e = self.edges[ei]\n",
    "        pushed = self.send_flow(source, node, min(f, e.cap - e.flow))\n",
    "        self.edges[ei].flow += pushed\n",
    "        self.edges[ei ^ 1].flow -= pushed\n",
    "        return pushed\n",
    "    def edmonds_karp(self, source, sink):\n",
    "        max_flow = 0\n",
    "        while self.bfs(source, sink):\n",
    "            if flow := self.send_flow(source, sink):\n",
    "                max_flow += flow\n",
    "            else:\n",
    "                break\n",
    "        return max_flow\n",
    "\n",
    "g = graph(5)\n",
    "g.add(0, 2, 100); g.add(0, 3, 50)\n",
    "g.add(2, 3, 50);  g.add(2, 4, 50); g.add(2, 1, 50)\n",
    "g.add(3, 4, 100)\n",
    "g.add(4, 1, 75)\n",
    "print(g.edmonds_karp(0, 1))\n",
    "\n",
    "g = graph(5)\n",
    "g.add(0, 2, 100); g.add(0, 3, 50)\n",
    "g.add(2, 4, 5); g.add(2, 1, 5)\n",
    "g.add(3, 4, 100)\n",
    "g.add(4, 1, 75)\n",
    "print(g.edmonds_karp(0, 1))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 41,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "125\n",
      "60\n"
     ]
    }
   ],
   "source": [
    "class graph:\n",
    "    class edge:\n",
    "        def __init__(self, node, cap, flow):\n",
    "            self.node, self.cap, self.flow = node, cap, flow\n",
    "        def __repr__(self):\n",
    "            return f'{self.node}:{self.cap}@{self.flow}'\n",
    "    def __init__(self, size):\n",
    "        self.nodes = [[] for _ in range(size)]\n",
    "        self.edges = []\n",
    "    def add(self, u, v, weight, directed = True):\n",
    "        if u == v: return\n",
    "        self.edges.append(graph.edge(v, weight, 0))\n",
    "        self.nodes[u].append(len(self.edges) - 1)\n",
    "        self.edges.append(graph.edge(u, weight if not directed else 0, 0))\n",
    "        self.nodes[v].append(len(self.edges) - 1)\n",
    "\n",
    "    # Find augmenting paths\n",
    "    def bfs(self, source, sink):\n",
    "        self.dist = [-1] * len(self.nodes)\n",
    "        self.prev = [(-1, -1)] * len(self.nodes)\n",
    "        queue = [source]\n",
    "        self.dist[source] = 0\n",
    "        while queue:\n",
    "            node = queue.pop(0)\n",
    "            if node == sink: break\n",
    "            for ei in self.nodes[node]:\n",
    "                e = self.edges[ei]\n",
    "                if e.cap - e.flow > 0 and self.dist[e.node] == -1:\n",
    "                    self.dist[e.node] = self.dist[node] + 1\n",
    "                    queue.append(e.node)\n",
    "                    self.prev[e.node] = (node, ei)\n",
    "        return self.dist[sink] != -1\n",
    "    \n",
    "    # Dinitz's algorithm\n",
    "    def dfs(self, source, sink, f = INF):\n",
    "        if source == sink or f == 0: return f\n",
    "        for i in range(self.last[source], len(self.nodes[source])):\n",
    "            e = self.edges[self.nodes[source][i]]\n",
    "            if self.dist[e.node] != self.dist[source] + 1: # not in layer graph!\n",
    "                continue\n",
    "            if pushed := self.dfs(e.node, sink, min(f, e.cap - e.flow)):\n",
    "                e.flow += pushed\n",
    "                self.edges[self.nodes[source][i] ^ 1].flow -= pushed\n",
    "                self.last[source] = i + 1\n",
    "                return pushed\n",
    "        self.last[source] = len(self.nodes[source])\n",
    "        return 0\n",
    "    def dinitz(self, source, sink):\n",
    "        max_flow = 0\n",
    "        while self.bfs(source, sink):\n",
    "            self.last = [0] * len(self.nodes)\n",
    "            while flow := self.dfs(source, sink):\n",
    "                max_flow += flow\n",
    "        return max_flow\n",
    "\n",
    "g = graph(5)\n",
    "g.add(0, 2, 100); g.add(0, 3, 50)\n",
    "g.add(2, 3, 50);  g.add(2, 4, 50); g.add(2, 1, 50)\n",
    "g.add(3, 4, 100)\n",
    "g.add(4, 1, 75)\n",
    "print(g.dinitz(0, 1))\n",
    "\n",
    "g = graph(5)\n",
    "g.add(0, 2, 100); g.add(0, 3, 50)\n",
    "g.add(2, 4, 5); g.add(2, 1, 5)\n",
    "g.add(3, 4, 100)\n",
    "g.add(4, 1, 75)\n",
    "print(g.dinitz(0, 1))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 38,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "(3, 407)"
      ]
     },
     "execution_count": 38,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "class graph:\n",
    "    class edge:\n",
    "        def __init__(self, node, cap, flow, cost):\n",
    "            self.node, self.cap, self.flow, self.cost = node, cap, flow, cost\n",
    "    def __init__(self, size):\n",
    "        self.nodes = [[] for _ in range(size)]\n",
    "        self.edges = []\n",
    "    def add(self, u, v, weight, cost, directed = True):\n",
    "        if u == v: return\n",
    "        self.edges.append(graph.edge(v, weight, 0, cost))\n",
    "        self.nodes[u].append(len(self.edges) - 1)\n",
    "        self.edges.append(graph.edge(u, weight if not directed else 0, 0, -cost))\n",
    "        self.nodes[v].append(len(self.edges) - 1)\n",
    "\n",
    "    # Shortest paths (Bellman-Ford)\n",
    "    def spfa(self, source, sink):\n",
    "        self.dist = [INF] * len(self.nodes)\n",
    "        queue = [source]\n",
    "        self.dist[source] = 0\n",
    "        self.visited[source] = True\n",
    "        while queue:\n",
    "            node = queue.pop(0)\n",
    "            self.visited[node] = False\n",
    "            for ei in self.nodes[node]:\n",
    "                e = self.edges[ei]\n",
    "                if e.cap - e.flow > 0 and self.dist[e.node] > self.dist[node] + e.cost:\n",
    "                    self.dist[e.node] = self.dist[node] + e.cost\n",
    "                    if not self.visited[e.node]:\n",
    "                        queue.append(e.node)\n",
    "                        self.visited[e.node] = True\n",
    "        return self.dist[sink] != INF\n",
    "\n",
    "    # Min-cost Max-flow algorithm\n",
    "    def dfs(self, source, sink, f = INF):\n",
    "        if source == sink or f == 0: return f\n",
    "        self.visited[source] = True\n",
    "        for i in range(self.last[source], len(self.nodes[source])):\n",
    "            e = self.edges[self.nodes[source][i]]\n",
    "            if self.visited[e.node] or self.dist[e.node] != self.dist[source] + e.cost: # not in layer graph!\n",
    "                continue\n",
    "            if pushed := self.dfs(e.node, sink, min(f, e.cap - e.flow)):\n",
    "                self.total_cost += pushed * e.cost\n",
    "                e.flow += pushed\n",
    "                self.edges[self.nodes[source][i] ^ 1].flow -= pushed\n",
    "                self.visited[source] = False\n",
    "                self.last[source] = i + 1\n",
    "                return pushed\n",
    "        self.visited[source] = False\n",
    "        self.last[source] = len(self.nodes[source])\n",
    "        return 0\n",
    "    def mcmf_dinitz(self, source, sink):\n",
    "        self.total_cost = 0\n",
    "        self.visited = [False] * len(self.nodes)\n",
    "        max_flow = 0\n",
    "        while self.spfa(source, sink):\n",
    "            self.last = [0] * len(self.nodes)\n",
    "            while flow := self.dfs(source, sink):\n",
    "                max_flow += flow\n",
    "        return max_flow, self.total_cost\n",
    "\n",
    "    # Model costs as bipartite graph\n",
    "    def assignment(prices):\n",
    "        nrows = len(prices)\n",
    "        ncols = len(prices[0])\n",
    "        g = graph(nrows + ncols + 2)\n",
    "        for r in range(nrows):\n",
    "            g.add(0, r + 1, 1, 0)\n",
    "        for r in range(nrows):\n",
    "            for c in range(ncols):\n",
    "                g.add(r + 1, nrows + c + 1, 1, prices[r][c])\n",
    "        for c in range(ncols):\n",
    "            g.add(nrows + c + 1, nrows + ncols + 1, 1, 0)\n",
    "        return g.mcmf_dinitz(0, nrows + ncols + 1)\n",
    "\n",
    "graph.assignment([\n",
    "    # j1   j2   j3\n",
    "    [108, 125, 150], # contractor 1\n",
    "    [150, 135, 175], # contractor 2\n",
    "    [122, 148, 250]  # contractor 3\n",
    "])"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  }
 ],
 "metadata": {
  "interpreter": {
   "hash": "eef2061d5c02a3b7277e3d8275f662ae1de2f1a336e2462c25d640b4e9ed1f20"
  },
  "kernelspec": {
   "display_name": "Python 3.10.1 64-bit ('3.10.1': pyenv)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.1"
  },
  "orig_nbformat": 4
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
